#!/usr/bin/python
# -*- encoding: utf-8 -*-
"""

"""
__author__ = 'Martin Martimeo <martin@martimeo.de>'
__date__ = '29.04.13 - 15:37'

from .api import ApiManager

__all__ = ['ApiManager']
__version__ = '0.4.4'
